<?php

/**
 * Viewdata/Videotex Server
 *
 * Inspired by Rob O'Donnell at irrelevant.com
 */

namespace App\Console\Commands;

use Illuminate\Console\Command;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Support\Facades\Log;
use Sock\{SocketClient,SocketServer,SocketException};

use App\Classes\Frame as FrameClass;
use App\Models\Frame as FrameModel;

class Server extends Command
{
	// Size of Bottom Line Pollution
	private $blp = 0;

	/**
	 * The name and signature of the console command.
	 *
	 * @var string
	 */
	protected $signature = 'server';

	/**
	 * The console command description.
	 *
	 * @var string
	 */
	protected $description = 'Start Videotex Server';

	/**
	 * Create a new command instance.
	 *
	 * @return void
	 */
	public function __construct()
	{
		parent::__construct();
	}

	/**
	 * Execute the console command.
	 *
	 * @return mixed
	 * @throws \Sock\SocketException
	 */
	public function handle()
	{
		/**
		 * Check dependencies
		 */
		if (! extension_loaded('sockets' ) ) {
			throw new SocketException(SocketException::CANT_ACCEPT,'Missing sockets extension');
		}

		if (! extension_loaded('pcntl' ) ) {
			throw new SocketException(SocketException::CANT_ACCEPT,'Missing pcntl extension');
		}

		define('MSG_TIMEWARP_ON', WHITE . 'TIMEWARP ON' . GREEN . 'VIEW INFO WITH *02');
		define('MSG_TIMEWARP_OFF', WHITE . 'TIMEWARP OFF' . GREEN . 'VIEWING DATE IS FIXED');
		define('MSG_TIMEWARP_TO', GREEN . 'TIMEWARP TO %s');
		define('MSG_TIMEWARP', WHITE . 'OTHER VERSIONS EXIST' . GREEN . 'KEY *02 TO VIEW');

		$server = new SocketServer(config('app.port'),config('app.bind'));
		$server->init();
		$server->setConnectionHandler([$this,'onConnect']);

		try {
			$server->listen();

		} catch (SocketException $e) {
			if ($e->getMessage() == 'Can\'t accept connections: "Success"')
				Log::debug('Server Terminated.');
			else
				Log::emergency('Uncaught Message: '.$e->getMessage());
		}
	}

	/**
	 * Connection handler
	 *
	 * @todo: Pressing * three times gets "Shouldnt get here".
	 */
	function onConnect(SocketClient $client) {
		Log::info('Connection from: ',['client'=>$client->getAddress()]);

		$pid = pcntl_fork();

		if ($pid == -1)
			throw new SocketException(SocketException::CANT_ACCEPT,'Could not fork process');

		// Parent return ready for next connection
		elseif ($pid)
			return;

		// We are now the child.
		try {
			$session_init = $session_option = FALSE;
			$session_note = '';         // TCP Session Notice
			$session_term = '';         // TCP Terminal Type

			$client->send(TCP_IAC . TCP_DO . TCP_OPT_SUP_GOAHEAD);     // DO SUPPRES GO AHEAD
			$client->send(TCP_IAC . TCP_WONT . TCP_OPT_LINEMODE);      // WONT LINEMODE
			$client->send(TCP_IAC . TCP_DO . TCP_OPT_ECHO);            // DO ECHO
//		    $client->send(TCP_IAC.TCP_AYT); // AYT
			$client->send(TCP_IAC . TCP_DO . TCP_OPT_TERMTYPE . TCP_IAC . TCP_SB . TCP_OPT_TERMTYPE . TCP_OPT_ECHO . TCP_IAC . TCP_SE); // Request Term Type

			$client->send(CLS);

			// Setup VARS
			$timewarp = FALSE;          // Is timewarp active.
			$timewarpalt = FALSE;       // Alternative timewarp frame to get
			$history = collect();       // Page history for going backwards
			$action = ACTION_GOTO;      // Initial action.
			$cmd = '';                  // Current *command being typed in
			$mode = FALSE;              // Current mode.

			$current = [];              // Attributes about the current page
										// field/fieldnum indexes are for fields on the active page
			$current['fieldreset'] = FALSE;          // Flag to reset position (used in fields)
			$current['fieldpos'] = 0;                // For current field, position within.
			$current['prevmode'] = FALSE;            // Previous mode - in case we need to go back to MODE_FIELD

			// @todo Get the login/start page, and if it is not available, throw the ERR_DATEBASE error.
			if (isset($config['loginpage'])) {
				$page = ['frame'=>$config['loginpage'],'index'=>'a'];
			} else if (!empty($service['start_page'])) {
				$page = ['frame'=>$service['start_page'],'index'=>'a'];
			} else {
				$page = ['frame'=>'98','index'=>'a'];            // next page
			}

			while ($action != ACTION_TERMINATE) {
				// Read a character from the client session
				$read = $client->read(1);

				// It appears that read will return '' instead of false when a disconnect has occurred.
				// We'll set it to NULL so its caught later
				if ($read === '')
					$read = NULL;

				if ($read != '') {
					dump(sprintf('Mode: [%s] CMD: [%s] frame: [%s] Received [%s (%s)]', $mode, $cmd, $page['frame'].$page['index'], $read, ord($read)));

					// Client initiation input
					// TELNET http://pcmicro.com/netfoss/telnet.html
					if ($read == TCP_IAC OR $session_init OR $session_option) {
						Log::debug(sprintf('Session Char (%s)', ord($read)), ['init' => $session_init, 'option' => $session_option]);

						switch ($read) {
							// Command being sent.
							case TCP_IAC:
								$session_init = TRUE;
								$session_note = 'IAC ';

								continue 2;

							case TCP_SB:
								$session_option = TRUE;

								continue 2;

							case TCP_SE:
								$session_option = $session_init = FALSE;
								Log::debug('Session Terminal: ' . $session_term);

								break;

							case TCP_DO:
								$session_note .= 'DO ';

								continue 2;

							case TCP_WILL:
								$session_note .= 'WILL ';

								continue 2;

							case TCP_WONT:
								$session_note .= 'WONT ';

								continue 2;

							case TCP_OPT_TERMTYPE:

								continue 2;

							case TCP_OPT_ECHO:
								$session_note .= 'ECHO';
								$session_init = FALSE;
								$read = '';
								Log::debug($session_note);

								continue;

							case TCP_OPT_SUP_GOAHEAD:
								$session_note .= 'SUPPRESS GO AHEAD';
								$session_init = FALSE;
								$read = '';
								Log::debug($session_note);

								continue;

							default:
								if ($session_option AND $read) {
									$session_term .= $read;
									$read = '';

								} else {
									Log::debug(sprintf('Unhandled char in session_init :%s (%s)', $read, ord($read)));
								}
						}
					}

					switch ($mode) {
						// Key presses during field input.
						case MODE_FIELD:
							dump(sprintf('** Processing Keypress in MODE_FIELD [%s (%s)]. Last POS [%s]',$read,ord($read),$current['fieldpos']));
							$cmd = '';
							$action = FALSE;

							switch ($fo->type()) {
								// Response frame.
								case 'a':
									switch ($read) {
										// End of field entry.
										case HASH:
											// Next Field
											$current['fieldnum']++;
											$current['fieldpos'] = 0;

											if ($current['fieldnum'] < $fo->fields->count()) {
												$current['fieldnum'] = $fo->getFieldId('edit',$current['fieldnum']);

												if ($current['fieldnum'] !== FALSE) {
													$current['field'] = $fo->getField($current['fieldnum']);
													$fielddata[$current['fieldnum']] = '';

													$client->send($this->outputPosition($current['field']->x,$current['field']->y).CON);
													$mode = MODE_FIELD;
													$fielddate[$current['fieldnum']] = '';

												// There were no (more) editable fields.
												} else {
													$action = ACTION_SUBMITRF;
												}

											} else {
												// Finished all editable fields.
												$action = ACTION_SUBMITRF;
											}

											break;

										case STAR:
											$current['prevmode'] = MODE_FIELD;
											$action = ACTION_STAR;
											$current['fieldpos'] = 0;
											$fielddata[$current['fieldnum']] = '';

											break;

										case KEY_LEFT:
											if ($current['fieldpos']--)
												$client->send(LEFT);

											break;

										case KEY_RIGHT:
											if ($current['fieldpos']++ < $current['field']->length)
												$client->send(RIGHT);

											break;

										case KEY_DOWN:
											if ($current['fieldpos'] + 40 < $current['field']->length) {
												$client->send(DOWN);
												$current['fieldpos'] = $current['fieldpos'] + 40;
											};

											break;

										case KEY_UP:
											if ($current['fieldpos'] - 40 >= 0) {
												$client->send($read);
												$current['fieldpos'] = $current['fieldpos'] - 40;
											};

											break;

										case CR:
											// On second or later line of a field
											if ($current['fieldpos'] + $current['field']->x > 40) {
												$client->send($read);
												$current['fieldpos'] = (($current['fieldpos'] + $current['field']->x) % 40) * 40;

											} else {
												$client->send($this->outputPosition($current['field']->x,$current['field']->y).CON);
												$current['fieldpos'] = 0;
											}

											break;

										case ESC:
											break;;

										// Record Data Entry
										default:
											if (ord($read) > 31 && $current['fieldpos'] < $current['field']->length) {
												$fielddata[$current['fieldnum']]{$current['fieldpos']} = $read;
												$current['fieldpos']++;
												$client->send($read);
											}
									}

									break;

								// Other Frame Types - Shouldnt get here.
								default:
									$client->close();

									throw new \Exception('Shouldnt get here', 500);

							}

							break;

						// Form submission: 1 to send, 2 not to send.
						case MODE_SUBMITRF:
							switch ($read) {
								// @todo Input received, process it.
								case '1':
									// dump(['line'=>__LINE__,'s' => $service, 'f' => $fielddata]);
									// @todo if send successful or not
									if (TRUE) {
										$this->sendBaseline($client,MSG_SENT);
										$mode = MODE_RFSENT;

									} else {
										$this->sendBaseline($client,ERR_NOTSENT);
										$mode = MODE_RFERROR;
									}

									break;

								case '2':
									$this->sendBaseline($client,MSG_NOTSENT);;
									$mode = MODE_RFNOTSENT;

									break;

								case STAR:
									$action = ACTION_STAR;

									break;
							}

							break;

						// Response form after Sent processing
						case MODE_RFSENT:
							$client->send(COFF);

							if ($read == HASH) {
								if (! empty($pagedata['route1'])) {
									$action = ACTION_GOTO;
									$page['frame'] = $pagedata['route1'];
									$page['index'] = 'a';

								} elseif (FrameModel::where('frame',$fo->frame())->where('index',$fo->index_next())->exists()) {
									$action = ACTION_GOTO;
									$page['frame'] = array_get($current,'page.frame');
									$page['index'] = chr(1 + ord($page['index']));

								} elseif (! empty($pagedata['route0'])) {
									$action = ACTION_GOTO;
									$page['frame'] = $pagedata['route0'];
									$page['index'] = 'a';

								// No further routes defined, go home.
								} else {
									$action = ACTION_GOTO;
									$page['frame'] = '0';
									$page['index'] = 'a';
								}

							} elseif ($read == STAR) {
								$action = ACTION_STAR;

								break;
							}

							break;

						// Response form after NOT sending
						case MODE_RFNOTSENT:

						// Response form ERROR
						case MODE_RFERROR:
							$client->send(COFF);

							if ($read == HASH) {
								if (! empty($pagedata['route2'])) {
									$action = ACTION_GOTO;
									$page['frame'] = $pagedata['route2'];
									$page['index'] = 'a';

								} elseif (FrameModel::where('frame',$fo->frame())->where('index',$fo->index_next())->exists()) {
									$action = ACTION_GOTO;
									$page['frame'] = $fo->frame();
									$page['index'] = $fo->index_next();

								} elseif (! empty($pagedata['route0'])) {
									$action = ACTION_GOTO;
									$page['frame'] = $pagedata['route0'];
									$page['index'] = 'a';

								// No further routes defined, go home.
								} else {
									$action = ACTION_GOTO;
									$page['frame'] = '0';
									$page['index'] = 'a';
								}

							} elseif ($read == STAR) {
								$action = ACTION_STAR;

								break;
							}

							break;

						// List of alternate frames has been presented
						case MODE_WARPTO:
							if (is_numeric($read) AND $read) {
								$timewarpalt = $alts->get($read-1)->id;
								$action = ACTION_GOTO;

							} elseif ($read === '0') {
								$action = ACTION_RELOAD;
							}

							break;

						// Not doing anything in particular.
						case MODE_COMPLETE:
						case FALSE:
							Log::debug('Idle', ['pid' => $pid]);
							$cmd = '';

							switch ($read) {
								case HASH:
									$action = ACTION_NEXT;
									break;

								case STAR:
									$action = ACTION_STAR;
									break;

								// Frame Routing
								case '0':
								case '1':
								case '2':
								case '3':
								case '4':
								case '5':
								case '6':
								case '7':
								case '8':
								case '9':
									if ($frame = $fo->route($read)) {
										$action = ACTION_GOTO;

									} else {
										$this->sendBaseline($client,ERR_ROUTE);
										$mode = $action = FALSE;
									}

									break;
							}

							break;

						// Currently accepting baseline input after a * was received
						case MODE_BL:
							echo "was waiting for page number\n";

							// if it's a number, continue entry
							if (strpos('0123456789', $read) !== FALSE) {
								$cmd .= $read;
								$client->send($read);
								$this->blp++;
							}

							// if we hit a special numeric command, deal with it.
							// Refresh page
							if ($cmd === '00') {
								$client->send(COFF);
								$action = ACTION_RELOAD;
								$cmd = '';

								break;
							}

							// Toggle Timewarp Mode
							if ($cmd === '01') {
								$client->send(COFF);
								$timewarp = !$timewarp;
								$this->sendBaseline($client,($timewarp ? MSG_TIMEWARP_ON : MSG_TIMEWARP_OFF));
								$cmd = '';
								$action = $mode = FALSE;

								break;
							}

							// Present Timewarp Frames
							if ($cmd === '02') {
								$client->send(COFF);
								$action = ACTION_INFO;
								$cmd = '';

								break;
							}

							// Reload page
							if ($cmd === '09') {
								$client->send(COFF);
								$action = ACTION_GOTO;
								$cmd = '';

								break;
							}

							// Another star aborts the command.
							if ($read === STAR) {
								$action = FALSE;
								$this->sendBaseline($client,'');
								$cmd = '';

								if ($current['prevmode'] == MODE_FIELD) {
									$mode = $current['prevmode'];
									$current['prevmode'] = FALSE;
									$client->send($this->outputPosition($current['field']->x,$current['field']->y).CON);
									$client->send(str_repeat('.', $current['field']->length));
									$current['fieldreset'] = TRUE;

								} else {
									$mode = false;
								}

								break;
							}

							// Complete request
							if ($read === HASH) {
								$client->send(COFF);
								$timewarpalt = FALSE;

								// Nothing typed between * and #
								if ($cmd === '') {
									$action = ACTION_BACKUP;

								// *# means go back
								} else {
									$page['frame'] = $cmd;
									$page['index'] = 'a';
									$action = ACTION_GOTO;
								}

								// Clear the command, we are finished processing
								$cmd = '';

								break;
							}

							break;

						default:
							Log::debug('Not sure what we were doing?');
					}
				}

				// This section performs some action if it is deemed necessary
				if ($action) {
					echo "Performing action $action\n";
				}

				switch ($action) {
					case ACTION_STAR:
						echo " star command started\n";
						$this->sendBaseline($client,GREEN.STAR,true);
						$client->send(CON);
						$action = false;
						$mode = MODE_BL;

						break;

					case ACTION_SUBMITRF:
						$action = false;
						$client->send(COFF);
						$this->sendBaseline($client,MSG_SENDORNOT);
						$mode = MODE_SUBMITRF;

						break;

					// GO Backwards
					case ACTION_BACKUP:
						// Do we have anywhere to go, drop the current page from the history.
						if ($history->count() > 1)
							$history->pop();

						$page = $history->last();

						Log::debug('Backing up to: '.$page['frame'].$page['index']);

					// Go to next index frame.
					case ACTION_NEXT:
						// We need this extra test in case we come from ACTION_BACKUP
						if ($action == ACTION_NEXT)
						{
							$current['page']['index'] = $fo->index();
							$page['index'] = $fo->index_next();
						}

					// Look for requested page
					case ACTION_GOTO:
						// If we wanted a "Searching..." message, this is where to put it.

						try {
							$fo = $timewarpalt
								? new FrameClass(FrameModel::findOrFail($timewarpalt))
								: (new FrameModel)->fetch($page['frame'],$page['index']);

						} catch (ModelNotFoundException $e) {
							$this->sendBaseline($client,ERR_PAGE);
							$mode = $action = FALSE;

							break;
						}

						// validate if we have access top it
						/*				if (isset($m['access']) && $m['access'] == 'n') {
											$this->sendBaseline($client,ERR_PAGE);
											$mode = $action = false;
											break;
										}
										if (isset($m['cug']) && is_numeric($m['cug']) && $m['cug'] && !in_array($m['cug'],$usercugs)) {
											$this->sendBaseline($client,ERR_PRIVATE);
											$mode = $action = false;
											break;
										}
						*/

						// we have access...
						//if ($r['varient_id'] != $varient['varient_id']) {
					/* if ($timewarpalt) {
							if (empty($v['varient_date'])) {
								$this->sendBaseline($client,sprintf(MSG_TIMEWARP_TO, 'unknown date'));
							} else {
								$this->sendBaseline($client,sprintf(MSG_TIMEWARP_TO, date('j f Y', strtotime($v['varient_date']))));
							}
							//$varient = array_merge($varient, array_intersect_key($r, $varient));
						$msg = "TIMEWARP";
							dump(['line'=>__LINE__,'write'=>'TIMEWARP MESSAGE']);
						}
					*/
						$current['page'] = $fo->page(TRUE);
						$current['fieldpos'] = 0;

						// Only if new location, not going backwards
						if (($history->last() != $page) AND ($action == ACTION_GOTO || $action == ACTION_NEXT)) {
							$history->push($page);
						}

					// drop into
					case ACTION_RELOAD:
						// @todo Move this $output into the object.
						if ($fo->hasFlag('clear'))
						{
							$this->blp = 0;
							$output = CLS;

						} else {
							$output = HOME
							;
							// Clear the baseline.
							$this->sendBaseline($client,'');
						}

						$output .= (string)$fo;

						if ($timewarpalt) {
							$this->sendBaseline($client,sprintf(MSG_TIMEWARP_TO,$fo->created() ? $fo->created()->format('Y-m-d H:i:s') : 'UNKNOWN'));
						}

						switch ($fo->type()) {
							default:
							// Standard Frame
							case 'i':
								$client->send($output);
								$mode = $action = false;
								break;

							// Active Frame. Prestel uses this for a Response Frame.
							case 'a':
								$client->send($output);
								// holds data entered by user.
								$fielddata = [];

								if (count($fo->fields)) {
									// Get our first editable field.
									$current['fieldnum'] = $fo->getFieldId('edit',0);
									$current['field'] = $fo->getField($current['fieldnum']);
									$current['fieldreset'] = TRUE;

									if ($current['fieldnum'] !== FALSE) {
										$mode = MODE_FIELD;

									// There were no editable fields.
									} else {
										$mode = MODE_COMPLETE;
										$client->send(COFF);
									}

									$current['fieldpos'] = 0;

								} else {
									$mode = FALSE;
								}

								break;

							// Terminate Frame
							case 't':
								$client->send($output);
								$action = ACTION_TERMINATE;
								break;

						}

						break;

					// Timewarp Mode
					case ACTION_INFO:
						$mode = $action = FALSE;
						$cmd = '';
						$y = 0;

						$output = $this->outputPosition(0, $y++) . WHITE . NEWBG . RED . 'TIMEWARP INFO FOR Pg.' . BLUE . $fo->page(). WHITE;
						//$output .= $this->outputPosition(0, $y++) . WHITE . NEWBG . BLUE . 'Service : ' . substr($service['service_name'] . str_repeat(' ', 27), 0, 27);
						//$output .= $this->outputPosition(0, $y++) . WHITE . NEWBG . BLUE . 'Varient : ' . substr($varient['varient_name'] . str_repeat(' ', 27), 0, 27);
						$output .= $this->outputPosition(0, $y++) . WHITE . NEWBG . BLUE . 'Dated   : ' .substr(($fo->created() ? $fo->created()->format('j F Y') : 'Unknown').str_repeat(' ', 27), 0, 27);

						$alts = $fo->alts()->get();

						if (count($alts) > 1) {
							$n = 1;

							$output .= $this->outputPosition(0, $y++) . WHITE . NEWBG . RED . 'ALTERNATIVE VERSIONS:' . str_repeat(' ', 16);

							foreach ($alts as $o) {
								$date = $o->created_at->format('d M Y');

								$line = WHITE.NEWBG;

								if ($timewarp) {
									$line .= RED.$n++;
								}

								$line .= BLUE.$date.' '.$o->note;

								$output .= $this->outputPosition(0,$y++).$line.str_repeat(' ',40-$this->strlenv($line));
							}

							if ($timewarp) {
								$mode = MODE_WARPTO;
							}

						}

						$client->send($output);

						break;
				}

				// We need to reposition the cursor to the current field.
				if ($current['fieldreset'] !== FALSE) {
					$client->send($this->outputPosition($current['field']->x,$current['field']->y).CON);
					$current['fieldreset'] = FALSE;
				}

				// Did the client disconnect
				if ($read === NULL || socket_last_error()) {
					$client->close();
					Log::debug('Client Disconnected: '.$client->getaddress());

					return FALSE;
				}
			}

		// Something bad happened. We'll log it and then disconnect.
		} catch (\Exception $e) {
			Log::emergency($e->getMessage());
		}

		$client->close();
		Log::debug('Disconnected: '.$client->getaddress());
	}

	/**
	 * Move the cursor via the shortest path.
	 */
	function outputPosition($x,$y) {
		// Take the shortest path.
		if ($y < 12) {
			return HOME.
				(($x < 21)
					? str_repeat(DOWN,$y).str_repeat(RIGHT,$x)
					: str_repeat(DOWN,$y+1).str_repeat(LEFT,40-$x));

		} else {
			return HOME.str_repeat(UP,24-$y).
				(($x < 21)
					? str_repeat(RIGHT,$x)
					: str_repeat(LEFT,40-$x));
		}
	}

	/**
	 * Send a message to the base line
	 *
	 * @note "$this->blp" remembers how many chars are there, so that they can be replace with next call
	 * @param $client
	 * @param $text
	 * @param bool $reposition
	 */
	function sendBaseline($client,$text,$reposition=FALSE) {
		$client->send(HOME.UP.$text.
			($this->blp > $this->strlenv($text)
				? str_repeat(' ',$this->blp-$this->strlenv($text)).
				($reposition ? HOME.UP.str_repeat(RIGHT,$this->strlenv($text)) : '')
				: '')
		);

		$this->blp = $this->strlenv($text);
	}

	/**
	 * Calculate the length of text
	 *
	 * ESC characters are two chars, and need to be counted as one.
	 *
	 * @param $text
	 * @return int
	 * @todo Implement this as a helper
	 */
	function strlenv($text):int {
		return strlen($text)-substr_count($text,ESC);
	}
}